#!/usr/bin/env python
# type: ignore

"""Tests for `access_py_telemetry` package."""

from access_py_telemetry.registry import TelemetryRegister
from pydantic import ValidationError
import pytest


def test_telemetry_register_unique(reset_telemetry_register):
    """
    Check that the TelemetryRegister class is a singleton & that we can register
    and deregister functions as we would expect.
    """
    TelemetryRegister._instances = {}
    session1 = TelemetryRegister("intake_catalog")
    session2 = TelemetryRegister("intake_catalog")

    # assert session1 is session2

    assert set(session1) >= {
        "esm_datastore.search",
        "DfFileCatalog.search",
        "DfFileCatalog.__getitem__",
    }

    session1.register("test_function")

    assert set(session2) >= {
        "esm_datastore.search",
        "DfFileCatalog.search",
        "DfFileCatalog.__getitem__",
        "test_function",
    }

    session1.deregister("test_function", "DfFileCatalog.__getitem__")

    session3 = TelemetryRegister("intake_catalog")

    assert set(session3) >= {
        "esm_datastore.search",
        "DfFileCatalog.search",
    }

    assert "test_function" not in session3
    assert "DfFileCatalog.__getitem__" not in session3

    assert set(session2) >= {
        "esm_datastore.search",
        "DfFileCatalog.search",
    }

    assert "test_function" not in session2
    assert "DfFileCatalog.__getitem__" not in session2


def test_telemetry_register_validation(reset_telemetry_register):
    session_register = TelemetryRegister("intake_catalog")

    with pytest.raises(ValidationError):
        session_register.register(1.0)

    with pytest.raises(ValidationError):
        session_register.deregister(1.0, 2.0, [3.0])

    def test_function():
        pass

    session_register.register(test_function)

    assert "test_function" in session_register

    for func_str in [
        "test_function",
        "esm_datastore.search",
        "DfFileCatalog.__getitem__",
        "DfFileCatalog.search",
    ]:
        assert func_str in str(session_register)
        assert func_str in repr(session_register)

    session_register.deregister(test_function)

    for func_str in [
        "esm_datastore.search",
        "DfFileCatalog.__getitem__",
        "DfFileCatalog.search",
    ]:
        assert func_str in str(session_register)
        assert func_str in repr(session_register)

    assert "test_function" not in session_register



# type: ignore
from pytest import fixture
from access_py_telemetry.api import ApiHandler, ProductionToggle
from access_py_telemetry.utils import ENDPOINTS
from access_py_telemetry.registry import TelemetryRegister


@fixture
def api_handler():
    """
    Get an instance of the APIHandler class, and then reset it after the test.

    """
    yield ApiHandler()

    ApiHandler._instance = None
    ApiHandler._server_url = "https://reporting.access-nri-store.cloud.edu.au"
    ApiHandler.headers = {service: {} for service in ENDPOINTS}
    ApiHandler.endpoints = {key: val for key, val in ENDPOINTS.items()}
    ApiHandler._extra_fields = {ep_name: {} for ep_name in ENDPOINTS.keys()}
    ApiHandler._pop_fields = {}


@fixture
def reset_telemetry_register():
    """
    Get the TelemetryRegister class for the catalog service.
    """
    yield TelemetryRegister
    TelemetryRegister._instances = {}


@fixture
def production_toggle():
    """
    Get the production toggle for the APIHandler class.
    """
    _ = ProductionToggle()
    del _
    yield ProductionToggle()
    ProductionToggle._instance = None
    ProductionToggle._production = True
    ProductionToggle.STAGING_URL = (
        "https://staging-reporting.access-nri-store.cloud.edu.au/api/"
    )
    ProductionToggle.PRODUCTION_URL = (
        "https://reporting.access-nri-store.cloud.edu.au/api/"
    )



# type: ignore

try:
    from access_py_telemetry import capture_registered_calls
    from IPython import get_ipython

    get_ipython().events.register("shell_initialized", capture_registered_calls)
    print("Intake telemetry extension loaded")
except ImportError as e:
    print("Intake telemetry extension not loaded")
    raise e



"""
Copyright 2022 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
SPDX-License-Identifier: Apache-2.0
"""

from functools import wraps
import sys
from typing import Any, Type, Iterable, Callable
import warnings
import platform
import uuid
import httpx
import asyncio
import pydantic
import re
import yaml
import multiprocessing
from pathlib import Path, PurePosixPath
import logging

from .utils import ENDPOINTS

logging.getLogger("httpx").setLevel(logging.WARNING)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

with open(Path(__file__).parent / "config.yaml", "r") as f:
    config = yaml.safe_load(f)

NRI_USER = True


class ProductionToggle:
    """
    Singleton class to hold info about whether the code is running in production
    or not.

    This class is a singleton so that the production status can be set once and
    accessed from anywhere in the code.

    Exposed functionality:
    - production: bool
        Whether the code is running in production or not. Setting this will also
        set the server URL to the production or staging URL.
    - debug: Callable
        A decorator that wraps a function in a try/except block. If the code is
        running in production, the function will be called normally. If the code
        is not running in production, the function will be called and any
        exceptions will be ignored. This is useful for debugging purposes.

    """

    _production = True
    _instance = None

    PRODUCTION_URL = "https://reporting.access-nri-store.cloud.edu.au/api/"
    STAGING_URL = "https://reporting-dev.access-nri-store.cloud.edu.au/api/"

    def __new__(cls: Type[Self]) -> Self:
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if hasattr(self, "initialized"):
            return None
        self.initialized = True

    @property
    def production(self) -> bool:
        return self._production

    @production.setter
    def production(self, prod: bool) -> None:
        """
        Set the production status.
        """
        if not isinstance(prod, bool):
            raise TypeError("Production status must be a boolean")
        if prod:
            ApiHandler().server_url = self.PRODUCTION_URL
        else:
            ApiHandler().server_url = self.STAGING_URL
        self._production = prod
        return None

    def debug(self) -> Callable[..., Any]:
        """
        Debugging decorator. Applying this to a function will wrap all telemetry
        calls in try/except blocks so that users never see any exceptions from
        the telemetry code.

        Notes
        -----
        We have to apply the branching logic *within* the decorator, because
        otherwise the logic gets applied at initialization time, and we can't
        change the production status after that.
        """

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if self.production:
                    try:
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            return func(*args, **kwargs)
                    except Exception:
                        return None
                else:
                    return func(*args, **kwargs)

            return wrapper

        return decorator

    def __str__(self) -> str:
        return f"ProductionToggle(production={self._production})"

    def __repr__(self) -> str:
        return f"ProductionToggle(production={self._production})"


TOGGLE = ProductionToggle()


class ApiHandler:
    """
    Singleton class to handle API requests. I'm only using a class here so we can save
    the extra_fields attribute.

    Singleton so that we can add extra fields elsewhere in the code and have them
    persist across all telemetry calls.

    To configure request timeouts and the multiprocessing context manually, configure
    the _request_timeout and _mproc_override class attributes as desired.
    """

    _instance = None
    endpoints = {service: endpoint for service, endpoint in ENDPOINTS.items()}
    headers: dict[str, dict[str, str]] = {service: {} for service in ENDPOINTS}
    _extra_fields: dict[str, dict[str, Any]] = {ep_name: {} for ep_name in ENDPOINTS}
    _pop_fields: dict[str, list[str]] = {}
    _request_timeout = None
    _mproc_override = None

    def __new__(cls: Type[Self], *args: Any, **kwargs: Any) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        server_url: str = "https://reporting.access-nri-store.cloud.edu.au",
    ) -> None:
        if hasattr(self, "_initialized"):
            return None
        self._initialized = True
        self._server_url = server_url

    @property
    def extra_fields(self) -> dict[str, Any]:
        return self._extra_fields

    @pydantic.validate_call
    def add_extra_fields(self, service_name: str, fields: dict[str, Any]) -> None:
        """
        Add an extra field to the telemetry data. Only works for services that
        already have an endpoint defined.
        """
        if service_name not in self.endpoints:
            raise KeyError(f"Endpoint for '{service_name}' not found")
        self._extra_fields[service_name] = fields
        return None

    @pydantic.validate_call
    def set_headers(
        self, service_names: str | Iterable[str] | None, headers: dict[str, str]
    ) -> None:
        """
        Add headers to the telemetry request for a given service or services, if
        specified.

        If service_names is None, the headers will be added to all services.
        """
        if isinstance(service_names, str):
            service_names = [service_names]

        for service_name in service_names or self.endpoints:
            if service_name not in self.endpoints:
                raise KeyError(f"Endpoint for '{service_name}' not found")
            self.headers[service_name] = headers
        return None

    @pydantic.validate_call
    def clear_headers(self, service_names: str | Iterable[str] | None = None) -> None:
        """
        Clear the headers for a given service or services, if specified.

        If service_names is None, the headers will be cleared for all services.
        """
        if isinstance(service_names, str):
            service_names = [service_names]

        for service_name in service_names or self.endpoints:
            if service_name not in self.endpoints:
                raise KeyError(f"Endpoint for '{service_name}' not found")
            self.headers[service_name] = {}
        return None

    @property
    def server_url(self) -> str:
        return self._server_url

    @server_url.setter
    def server_url(self, url: str) -> None:
        """
        Set the server URL for the telemetry API.
        """
        if NRI_USER and (
            "https://reporting-dev.access-nri-store.cloud.edu.au/" not in url
            and "https://reporting.access-nri-store.cloud.edu.au/" not in url
        ):
            warnings.warn(
                "Server URL not an ACCESS-NRI Reporting API URL",
                stacklevel=2,
                category=UserWarning,
            )
        if NRI_USER and not url.lower().endswith(("api", "api/")):
            warnings.warn(
                "Server URL does not end with 'api' or 'api/' - this is likely an error",
                stacklevel=2,
                category=UserWarning,
            )
        self._server_url = url
        return None

    @property
    def pop_fields(self) -> dict[str, list[str]]:
        return self._pop_fields

    @property
    def request_timeout(self) -> float | None:
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, timeout: float | None) -> None:
        """
        Set the request timeout for the telemetry API.
        """
        if timeout is None:
            self._request_timeout = None
            return None
        if not isinstance(timeout, (int, float)):
            raise TypeError("Timeout must be a number")
        elif timeout <= 0 or not isinstance(timeout, (int, float)):
            raise ValueError("Timeout must be a positive number")

        self._request_timeout = timeout
        return None

    @pydantic.validate_call
    def remove_fields(self, service: str, fields: str | Iterable[str]) -> None:
        """
        Set the fields to remove from the telemetry data for a given service. Useful for excluding default
        fields that are not needed for a particular telemetry call: eg, removing
        Session tracking if a CLI is being used.

        Note: This does not use a set union, so you must specify all fields you want to remove in one call.
        # TODO: Maybe make this easier to use?
        """
        if isinstance(fields, str):
            fields = [fields]
        self._pop_fields[service] = list(fields)

    @TOGGLE.debug()
    def send_api_request(
        self,
        service_name: str,
        function_name: str,
        args: list[Any] | tuple[Any, ...],
        kwargs: dict[str, str | Any],
    ) -> None:
        """
        Send an API request with telemetry data.

        Parameters
        ----------
        service_name : str
            The name of the service to send the telemetry data to.
        function_name : str
            The name of the function being tracked.
        args : list
            The list of positional arguments passed to the function.
        kwargs : dict
            The dictionary of keyword arguments passed to the function.

        Returns
        -------
        None

        Warnings
        --------
        RuntimeWarning
            If the request fails.

        """

        telemetry_data = self._create_telemetry_record(
            service_name, function_name, args, kwargs
        )

        endpoint = self._get_endpoints(service_name)

        telemetry_headers = self.headers.get(service_name, {})

        endpoint = _format_endpoint(self.server_url, endpoint)

        send_in_loop(
            endpoint,
            telemetry_data,
            telemetry_headers,
            self._request_timeout,
            self._mproc_override,
        )
        return None

    def _get_endpoints(self, service_name: str) -> str:
        """
        Get the endpoint for a given service name.
        """
        try:
            endpoint = self.endpoints[service_name]
        except KeyError as e:
            raise KeyError(
                f"Endpoint for '{service_name}' not found in {self.endpoints}"
            ) from e
        return endpoint

    def _create_telemetry_record(
        self,
        service_name: str,
        function_name: str,
        args: list[Any] | tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """
        Create and return a telemetry record, cache it as an instance attribute.


        Notes
        -----
        SessionID() is a lazily evaluated singleton, so it looks like we are
        going to generate a new session ID every time we call this function, but we
        aren't. I've also modified __get__, so SessionID() evaluates to a string.
        """
        telemetry_data = {
            # "name": getpass.getuser(), # Until we work out the privacy policy nightmare
            "function": function_name,
            "args": args,
            "kwargs": kwargs,
            "session_id": SessionID(),
            **self.extra_fields.get(service_name, {}),
        }

        for field in self.pop_fields.get(service_name, []):
            telemetry_data.pop(field)

        self._last_record = telemetry_data
        return telemetry_data


class SessionID:
    """
    Singleton class to store and generate a unique session ID.

    This class ensures that only one instance of the session ID exists. The session
    ID is generated the first time it is accessed and is represented as a string.
    The session ID is created using using the UUID4 algorithm.

    Methods:
        __new__(cls, *args, **kwargs): Ensures only one instance of the class is created.
        __init__(self): Initializes the instance.
        __get__(self, obj: object, objtype: type | None = None) -> str: Generates and returns the session ID.
        create_session_id() -> str: Static method to create a unique session ID.
    """

    _instance = None

    def __new__(cls: type[Self]) -> Self:
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if hasattr(self, "initialized"):
            return None
        self.initialized = True

    def __get__(self, obj: object, objtype: type | None = None) -> str:
        if not hasattr(self, "value"):
            self.value = SessionID.create_session_id()
        return self.value

    @staticmethod
    def create_session_id() -> str:
        """
        Generate a unique session ID.
        """
        return str(uuid.uuid4())


async def send_telemetry(
    endpoint: str,
    data: dict[str, Any],
    headers: dict[str, str],
    warn: bool | None = None,
) -> None:
    """
    Asynchronously send telemetry data to the specified endpoint.

    Parameters
    ----------
    endpoint : str
        The URL to send the telemetry data to.
    data : dict
        The telemetry data to send.
    headers : dict
        The headers to send the telemetry data with.
    warn : bool, optional
        If True, a warning will be raised if the request fails. If False, no
        warning will be raised. If None, warn will default the value of
        ` not ProductionToggle().production`. It wil also enable some status info
        about the request being sent.

    Returns
    -------
    None

    Warnings
    --------
    RuntimeWarning
        If the request fails.
    """
    if warn is None:
        warn = not ProductionToggle().production

    headers = {
        "Content-Type": "application/json",
        **headers,
    }
    async with httpx.AsyncClient() as client:
        try:
            if warn:
                print(f"Posting telemetry to {endpoint}")
            response = await client.post(endpoint, json=data, headers=headers)
            response.raise_for_status()
        except (httpx.RequestError, httpx.HTTPStatusError) as e:
            if warn:
                warnings.warn(
                    f"Request failed: {e}", category=RuntimeWarning, stacklevel=2
                )
    return None


def send_in_loop(
    endpoint: str,
    telemetry_data: dict[str, Any],
    telemetry_headers: dict[str, str] | None = None,
    timeout: float | None = None,
    mproc_override: str | None = None,
) -> None:
    """
    Wraps the send_telemetry function in an event loop. This function will:
    - Check if an event loop is already running
    - If an event loop is running, send the telemetry data in the background
    - If an event loop is not running, create a new event loop in a separate process
        and send the telemetry data in the background using that loop.

    Parameters
    ----------
    endpoint : str
        The URL to send the telemetry data to.
    telemetry_data : dict
        The telemetry data to send.
    headers : dict, optional
        The headers to send the telemetry data with.
    timeout : float, optional
        The maximum time to wait for the coroutine to finish. If the coroutine takes
        longer than this time, a TimeoutError will be raised. If None, the coroutine
        will terminate after 60 seconds. Timeout will also revert to 60 seconds if
        set to 0.
    mproc_override : str, optional
        The multiprocessing context to use. If None, the context will be set to "fork"
        on Linux systems and "spawn" on Windows/ MacOS systems. If a context is specified,
        it will be used regardless of the system.

    Returns
    -------
    None

    """
    timeout = timeout or 60

    telemetry_headers = telemetry_headers or {}

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        _run_in_proc(
            endpoint, telemetry_data, telemetry_headers, timeout, mproc_override
        )
    else:
        loop.create_task(send_telemetry(endpoint, telemetry_data, telemetry_headers))
        return None


def _run_event_loop(
    endpoint: str,
    telemetry_data: dict[str, Any],
    telemetry_headers: dict[str, str] | None = None,
    warn: bool | None = None,
) -> None:
    """
    Handles the creation and running of an event loop for sending telemetry data.
    This function is intended to be run in a separate process, and will:
    - Create a new event loop
    - Send the telemetry data
    - Run the event loop until the telemetry data is sent

    Parameters
    ----------
    endpoint : str
        The URL to send the telemetry data to.
    telemetry_data : dict
        The telemetry data to send.
    telemetry_headers : dict, optional
        The headers to send the telemetry data with.
    warn : bool, optional
        If True, a warning will be raised if the request fails. If False, no
        warning will be raised. If None, warn will default the value of
        ` not ProductionToggle().production`. It will also enable some status info
        about the request being sent.

    Returns
    -------
    None

    Notes
    -----
    We pass through warn here as otherwise ProductionToggle() will be initialized
    in the main process, and we want to avoid that.
    """
    if warn is None:
        warn = not ProductionToggle().production
    telemetry_headers = telemetry_headers or {}
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(
        send_telemetry(endpoint, telemetry_data, telemetry_headers, warn)
    )


def _run_in_proc(
    endpoint: str,
    telemetry_data: dict[str, Any],
    telemetry_headers: dict[str, str] | None,
    timeout: float = 60,
    mproc_override: str | None = None,
) -> None:
    """
    Handles the creation and running of a separate process for sending telemetry data.
    This function will:
    - Create a new process and run the _run_event_loop function in that process
    - Wait for the process to finish
    - If the process takes longer than the specified timeout, terminate the process
        and raise a warning

    Parameters
    ----------
    endpoint : str
        The URL to send the telemetry data to.
    telemetry_data : dict
        The telemetry data to send.
    timeout : float
        The maximum time to wait for the process to finish.
    telemetry_headers : dict, optional
        The headers to send the telemetry data with.
    mproc_override : str, optional
        The multiprocessing context to use. If None, the context will be set to "fork"
        on Linux systems and "spawn" on Windows systems. If a context is specified, it
        will be used regardless of the system.

    Returns
    -------
    None

    """
    telemetry_headers = telemetry_headers or {}

    if not mproc_override:
        ctx_type = "fork" if platform.system().lower() == "linux" else "spawn"
    else:
        ctx_type = mproc_override

    # Mypy gets upset below because it doesn't know we wont use "fork" on Windows
    proc = multiprocessing.get_context(ctx_type).Process(  # type: ignore
        target=_run_event_loop,
        args=(endpoint, telemetry_data, telemetry_headers),
    )
    proc.start()
    proc.join(timeout)
    if proc.is_alive():
        proc.terminate()
        warnings.warn(
            f"Telemetry data not sent within {timeout} seconds",
            category=RuntimeWarning,
            stacklevel=2,
        )
    return None


def _format_endpoint(server_url: str, endpoint: str) -> str:
    """
    Concatenates the server URL and endpoint, ensuring that there is only one
    slash between them.
    """
    endpoint = str(PurePosixPath(server_url) / endpoint.lstrip("/"))
    return re.sub(r"^(https?:/)(.*?)(?<!/)\/?$", r"\1/\2/", endpoint)



"""Console script for access_py_telemetry."""

import access_py_telemetry
from typing import Sequence
from shutil import copy2
import argparse
import filecmp
from pathlib import Path

PACKAGE_ROOT = Path(access_py_telemetry.__file__).parent


def configure_telemetry(argv: Sequence[str] | None = None) -> int:
    """Console script for configuring ipython telemetry."""
    parser = argparse.ArgumentParser(description="Configure ipython telemetry.")
    parser.add_argument("--disable", action="store_true", help="Disable telemetry.")
    parser.add_argument("--enable", action="store_true", help="Enable telemetry.")
    parser.add_argument("--status", action="store_true", help="Check telemetry status.")
    parser.add_argument("--silent", action="store_true", help="Suppress output.")

    HOME = Path.home()
    telemetry_fname = HOME / ".ipython" / "profile_default" / "startup" / "telemetry.py"
    template_file = PACKAGE_ROOT / "templates" / "telemetry_template.py"
    telem_file_exists = telemetry_fname.exists()

    args = parser.parse_args(argv)

    arg_dict = {
        "disable": args.disable,
        "enable": args.enable,
        "status": args.status,
    }

    if not any(arg_dict.values()):
        parser.print_help()
        return 0

    if len([arg for arg in arg_dict.values() if arg]) > 1:
        print("Only one of --disable, --enable, or --status can be used at a time.")
        return 1

    if args.status:
        if telem_file_exists and filecmp.cmp(telemetry_fname, template_file):
            print("Telemetry enabled.") if not args.silent else None
        elif telem_file_exists and not filecmp.cmp(telemetry_fname, template_file):
            (
                print(
                    "Telemetry enabled but misconfigured. Run `access-py-telemetry --disable && access-py-telemetry --enable` to fix."
                )
                if not args.silent
                else None
            )
        else:
            print("Telemetry disabled.") if not args.silent else None
        return 0

    if args.disable:
        if telem_file_exists:
            telemetry_fname.unlink()
            print("Telemetry disabled.") if not args.silent else None
        else:
            print("Telemetry already disabled.") if not args.silent else None
        return 0

    if args.enable:
        if telem_file_exists:
            print("Telemetry already enabled.") if not args.silent else None
            return 0

        if not telemetry_fname.parent.exists():
            telemetry_fname.parent.mkdir(parents=True)
        copy2(template_file, telemetry_fname)
        print("Telemetry enabled.") if not args.silent else None

    return 0



"""
Copyright 2022 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
SPDX-License-Identifier: Apache-2.0

Top-level package for access-ipy-telemetry.
"""

import warnings
from IPython.core.getipython import get_ipython
from IPython.core.interactiveshell import InteractiveShell

from . import _version
from .ast import capture_registered_calls
from .api import SessionID  # noqa
from .registry import RegisterWarning
from .utils import ENDPOINTS, REGISTRIES


# Make sure that our registries & endpoints match up
if not set(REGISTRIES.keys()) == set(ENDPOINTS.keys()):
    warnings.warn(
        "Mismatch between registries and endpoints - some telemetry calls may not be recorded.",
        category=RegisterWarning,
    )


def load_ipython_extension(ipython: InteractiveShell) -> None:
    """
    Load the IPython extension and register it to run before cells.
    """
    ipython.events.register("pre_run_cell", capture_registered_calls)  # type: ignore
    return None


# Register the extension
ip = get_ipython()  # type: ignore
if ip:
    load_ipython_extension(ip)


__version__ = _version.get_versions()["version"]  # type: ignore



#!/usr/bin/env python
# type: ignore

"""Tests for the AST module"""

import ast
import sys
import pytest
from access_py_telemetry.ast import CallListener, strip_magic, capture_registered_calls
from unittest.mock import MagicMock


class MockInfo:
    def __init__(self, raw_cell=None):
        self.raw_cell = raw_cell


def test_ast_instance_method():
    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    def func(self):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass

instance = MyClass()
mycall = instance.func()

instance.uncaught_func()
"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["MyClass.func"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "MyClass.func",
    }

    assert "MyClass.uncaught_func" not in visitor._caught_calls


def test_ast_bare_function():
    mock_info = MockInfo()
    mock_info.raw_cell = """
def registered_func():
    return None

def unregistered_func():
    return None

registered_func()
unregistered_func()

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["registered_func"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "registered_func",
    }

    assert "uncaught_func" not in visitor._caught_calls


def test_ast_unparse_bare_function():
    mock_info = MockInfo()
    mock_info.raw_cell = """

import pandas as pd

def registered_func():
    return None

def registered_func2(x):
    return None


def unregistered_func():
    return None

registered_func()
unregistered_func()

registered_func2(pd.DataFrame())

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["registered_func", "registered_func2"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "registered_func",
        "registered_func2",
    }

    assert "uncaught_func" not in visitor._caught_calls


@pytest.mark.xfail
def test_ast_aliased_function():
    """
    This will require more sophisticated analysis to catch aliased functions. Maybe
    we can look at this eventually
    """
    mock_info = MockInfo()
    mock_info.raw_cell = """
def registered_func():
    return None

def unregistered_func():
    return None


reg_func = registered_func

reg_func()

unregistered_func()

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["registered_func"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "registered_func",
    }

    assert "uncaught_func" not in visitor._caught_calls


@pytest.mark.xfail
def test_ast_instantiate_and_call():
    """
    Need to figure out how to catch the instantiation of a class and then call a method
    on it. Not needed yet
    """
    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    def func(self):
        self.set_var = set()


MyClass().func()

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["MyClass.func"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "MyClass.func",
    }


@pytest.mark.xfail
def test_ast_class_method():
    """
    Class methods don't work with the CallListener yet
    """
    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    @classmethod
    def class_func(cls):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass


MyClass.func(instance)

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["MyClass.class_func"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "MyClass.class_func",
    }


def test_ast_indexing():
    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    def func(self):
        self.set_var = set()

    def __getitem__(self, key):
        return [1, 2, 3]

instance = MyClass()
mycall = instance['some_item']

l = [1, 2, 3]

l[0]

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["MyClass.__getitem__", "list.__getitem__"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "MyClass.__getitem__",
        "list.__getitem__",
    }


def test_ast_nested_function():
    mock_info = MockInfo()
    mock_info.raw_cell = """
import os

os.path.join("some","paths")

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["os.path.join"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "os.path.join",
    }


def test_ast_aliased_module():
    mock_info = MockInfo()
    mock_info.raw_cell = """
import os as operating_system

operating_system.path.join("some","paths")

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["os.path.join"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "os.path.join",
    }


@pytest.mark.parametrize(
    "raw_cell, called_with",
    [
        (
            """
class MyClass:
    def func(self):
        self.set_var = set()

    def __getitem__(self, key):
        return [1, 2, 3]

instance = MyClass()

search_str = 'some_item'

mycall = instance[search_str] 
""",
            ("mock", "MyClass.__getitem__", ["some_item"], {}),
        ),
        (
            """
class MyClass:
    def func(self):
        self.set_var = set()

    def __getitem__(self, key):
        return [1, 2, 3]

instance = MyClass()

mycall = instance['directly_used_string'] 
""",
            ("mock", "MyClass.__getitem__", ["'directly_used_string'"], {}),
        ),
    ],
)
def test_ast_aliased_index(raw_cell, called_with):
    """
    We need to make sure that we properly catch, eg.
    ```python
    experiment_name = 'my_expt'
    esm_ds = catalog[experiment_name]
    ```
    and record the call to catalog.__getitem_ with an argument of 'my_expt'
    rather than the string identifier of the variable holding it, `experiment_name`
    """

    mock_info = MockInfo()
    mock_info.raw_cell = raw_cell
    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["MyClass.__getitem__", "list.__getitem__"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    mock_api_handler.send_api_request.assert_called_once_with(*called_with)


def test_import_catalog():
    mock_info = MockInfo()
    mock_info.raw_cell = """
import intake
intake.cat.access_nri

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["intake.cat.access_nri"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "intake.cat.access_nri",
    }


def test_import_assign_catalog():
    mock_info = MockInfo()
    mock_info.raw_cell = """
import intake
cat = intake.cat.access_nri

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["intake.cat.access_nri"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "intake.cat.access_nri",
    }


@pytest.mark.xfail
def test_import_catalog_traverse_imports():
    """
    This fails because the CallListener doesn't traverse the imports yet. We can
    do this using importlib and then parsing what importlib imports too I think,
    but let's save that for another day
    """
    mock_info = MockInfo()
    mock_info.raw_cell = """
import intake 
intake.cat.access_nri
"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["intake.catalog.Catalog.__init__"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {
        "intake.cat.access_nri",
    }


def test_match_ipython_magic():
    """
    This test is to check that the IPython magic commands are not caught by the
    CallListener. This is important because we don't want to send telemetry for
    IPython magic commands.
    """
    mock_info = MockInfo()
    mock_info.raw_cell = r"""
!ls
%%timeit
class MyClass:
    @classmethod
    def class_func(cls):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass

MyClass.func(instance)

MyClass.func??

MyClass.func?
    """

    python_code = r"""
class MyClass:
    @classmethod
    def class_func(cls):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass

MyClass.func(instance)


    """

    parsed_w_magic = strip_magic(mock_info.raw_cell)
    parsed_wo_magic = strip_magic(python_code)

    assert parsed_w_magic == parsed_wo_magic

    a = ast.dump(ast.parse(parsed_w_magic), annotate_fields=False)
    b = ast.dump(ast.parse(parsed_wo_magic), annotate_fields=False)
    assert a == b


def test_parse_invalid_code():
    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    def func(self):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass

instance = MyClass()
mycall = instance.func()

    instance.uncaught_func()


"""

    capture_registered_calls(mock_info)

    mock_info = MockInfo()
    mock_info.raw_cell = """
class MyClass:
    def func(self):
        self.set_var = set()

    def uncaught_func(self, *args, **kwargs):
        pass

@instance = MyClass()
1mycall = instance.func()

    instance.uncaught_func()


"""

    capture_registered_calls(mock_info)


def test_implicit_boolean_conversion():
    mock_info = MockInfo()
    mock_info.raw_cell = """

import numpy as np

arr = np.array([0, 2, 3])

arr.mean()

"""

    f = sys._getframe()
    exec(mock_info.raw_cell, globals(), f.f_locals)
    mock_user_ns = f.f_locals

    mock_registry = {"mock": ["ndarray.mean"]}

    mock_api_handler = MagicMock()

    tree = ast.parse(mock_info.raw_cell)

    visitor = CallListener(mock_user_ns, mock_registry, mock_api_handler)

    visitor.visit(tree)

    assert visitor._caught_calls == {"ndarray.mean"}
